@@ -1,11 +1,13 @@
+import contextlib
 import dataclasses
-import functools
 import uuid
 
 import json
 import logging
 import os
-from typing import TextIO, Literal, Dict
+from contextlib import contextmanager, asynccontextmanager
+from contextvars import ContextVar
+from typing import Literal, AsyncContextManager, Protocol, AsyncGenerator, Callable
 from functools import cached_property
 
 from openai._types import NOT_GIVEN, NotGiven
@@ -16,19 +18,23 @@ from openai import AsyncOpenAI
 
 from openai.types.chat import ChatCompletion, ChatCompletionMessage, ChatCompletionMessageToolCall, \
     ChatCompletionToolMessageParam, ChatCompletionSystemMessageParam, ChatCompletionUserMessageParam, \
-    ChatCompletionToolChoiceOptionParam, ChatCompletionAssistantMessageParam, ChatCompletionMessageParam, \
-    ChatCompletionToolParam
+    ChatCompletionToolChoiceOptionParam, ChatCompletionAssistantMessageParam, \
+    ChatCompletionToolParam, ChatCompletionNamedToolChoiceParam, ParsedChatCompletion, ParsedChatCompletionMessage
 from openai.types.completion_usage import CompletionUsage
-from functioncalming.utils import invoke_callback_function, FineTuningData, InnerValidationError, \
-    create_openai_function, OpenAIFunction, ToolCallError, create_abbreviated_openai_function, \
-    serialize_openai_function_result, log_finetuning_data
-from functioncalming.types import BaseModelOrJsonCompatible, Messages, BaseModelFunction
 
-_client = None
+from functioncalming.context import set_calm_context
+from functioncalming.utils import InnerValidationError, \
+    create_openai_function, OpenAIFunction, ToolCallError, create_abbreviated_openai_function, \
+    serialize_openai_function_result, rebuild_cached_models
+from functioncalming.types import Messages, JsonCompatibleFunction, JsonCompatible
+
+USING_STRUCTURED_OUTPUTS = "Using Structured Outputs without tool calling for this request."
+
+_openai_client: ContextVar[AsyncOpenAI | None] = ContextVar('_openai_client', default=None)
 
 
 def get_client():
-    global _client
+    _client = _openai_client.get()
     if not _client:
         _client = AsyncOpenAI(
             api_key=os.environ.get("OPENAI_API_KEY"),
@@ -37,39 +43,113 @@ def get_client():
         )
     return _client
 
+@contextmanager
+def set_openai_client(client: AsyncOpenAI):
+    token = _openai_client.set(client)
+    yield
+    _openai_client.reset(token)
 
-def set_client(client: AsyncOpenAI):
-    global _client
-    _client = client
+type DefaultBehavior = Literal['default_behavior']
+DEFAULT_BEHAVIOR: DefaultBehavior = "default_behavior"
+
+def register_model(
+        model_name: str,
+        supports_structured_outputs: bool,
+        cost_per_1mm_input_tokens: float,
+        cost_per_1mm_output_tokens: float
+):
+    if supports_structured_outputs:
+        STRUCTURED_OUTPUTS_SUPPORTED.add(model_name)
+    COSTS_BY_MODEL[model_name] = (cost_per_1mm_input_tokens, cost_per_1mm_output_tokens)
+
+# for registering new models externally when I forget to update the library again
+STRUCTURED_OUTPUTS_SUPPORTED = set()
 
 
-DEFAULT_BEHAVIOR = "default_behavior"
+def structured_outputs_available(model_name: str):
+    if model_name == "gpt-4o-2024-05-13":
+        return False
+    if model_name in STRUCTURED_OUTPUTS_SUPPORTED:
+        return True
+    return model_name.startswith('o1') or model_name.startswith('o3') or model_name.startswith('gpt-4o')
 
 
+# cost per 1MM token
 COSTS_BY_MODEL = {
+    # o1 and 01-mini
+    "o1-preview": (15.0, 60.0),
+    "o1-preview-2024-09-12": (15.0, 60.0),
+    "o1-mini": (3.0, 12.0),
+    "o1-mini-2024-09-12": (3.0, 12.0),
+    # 4o and 4o-mini
+    "gpt-4o-2024-05-13": (5.0, 15.0),  # no structured outputs yet!
+    "gpt-4o-2024-08-06": (2.5, 10.0),
+    "gpt-4o": (2.5, 10.0),
+    "gpt-4o-mini": (0.15, 0.6),
+    "gpt-4o-mini-2024-07-18": (0.15, 0.6),
     # GPT-4 Turbo
-    "gpt-4-turbo": (0.01, 0.03),
-    "gpt-4-0125-preview": (0.01, 0.03),  # actually a Turbo model
-    "gpt-4-1106-preview": (0.01, 0.03),  # actually a Turbo model
-    "gpt-4-vision-preview": (0.01, 0.03),
-    "gpt-4-1106-vision-preview": (0.01, 0.03),
+    "gpt-4-turbo": (10., 30.),
+    "gpt-4-0125-preview": (10., 30.),  # actually a Turbo model
+    "gpt-4-1106-preview": (10., 30.),  # actually a Turbo model
+    "gpt-4-vision-preview": (10., 30.),
+    "gpt-4-1106-vision-preview": (10., 30.),
+    "gpt-4-turbo-2024-04-09": (10., 30.),
     # GPT-4
-    "gpt-4": (0.03, 0.06),
-    "gpt-4-0613": (0.03, 0.06),
-    "gpt-4-32k": (0.06, 0.12),
+    "gpt-4": (30., 60.),
+    "gpt-4-0613": (30., 60.),
+    "gpt-4-32k": (60., 120.),
     "gpt-4-32k-0613": (0.06, 0.12),
     # GPT 3.5
-    "gpt-3.5-turbo": (0.0005, 0.0015),
-    "gpt-3.5-turbo-0125": (0.0005, 0.0015),
-    "gpt-3.5-turbo-1106": (0.0005, 0.0015),
-    "gpt-3.5-turbo-0301": (0.0005, 0.0015),  # available in playground but not in docs?
-    "gpt-3.5-turbo-instruct": (0.0015, 0.0020),  # not actually a chat model
-    # deprecated 3.5
-    "gpt-3.5-turbo-0613": (0.0005, 0.0015),
-    "gpt-3.5-turbo-16k-0613": (0.0005, 0.0015),
-    "gpt-3.5-turbo-16k": (0.0005, 0.0015),
+    "gpt-3.5-turbo": (0.5, 1.5),  # not sure actually
+    "gpt-3.5-turbo-0125": (0.5, 1.5),
+    "gpt-3.5-turbo-1106": (1., 2.),
+    "gpt-3.5-turbo-instruct": (1.5, 2.0),  # not actually a chat model
+    "gpt-3.5-turbo-16k-0613": (3., 4.),
+    "gpt-3.5-turbo-0613": (1.5, 2.),
+    "gpt-3.5-turbo-0301": (1.5, 2.),
+    "gpt-3.5-turbo-16k": (3., 4.),
 }
 
+# TODO refactor these
+for model_name in ('gpt-4.1', 'gpt-4.1-2025-04-14'):
+    register_model(
+        model_name=model_name,
+        supports_structured_outputs=True,
+        cost_per_1mm_input_tokens=2.0,
+        cost_per_1mm_output_tokens=8.0,
+    )
+for model_name in ('gpt-4.1-mini', 'gpt-4.1-mini-2025-04-14'):
+    register_model(
+        model_name=model_name,
+        supports_structured_outputs=True,
+        cost_per_1mm_input_tokens=0.4,
+        cost_per_1mm_output_tokens=1.6,
+    )
+
+for model_name in ('gpt-4.1-nano', 'gpt-4.1-nano-2025-04-14'):
+    register_model(
+        model_name=model_name,
+        supports_structured_outputs=True,
+        cost_per_1mm_input_tokens=0.1,
+        cost_per_1mm_output_tokens=0.4,
+    )
+
+for model_name in ('gpt-4o', 'gpt-4o-2024-08-06'):
+    register_model(
+        model_name=model_name,
+        supports_structured_outputs=True,
+        cost_per_1mm_input_tokens=2.5,
+        cost_per_1mm_output_tokens=10.0,
+    )
+
+for model_name in ('gpt-4o-mini', 'gpt-4o-mini-2024-07-18'):
+    register_model(
+        model_name=model_name,
+        supports_structured_outputs=True,
+        cost_per_1mm_input_tokens=.15,
+        cost_per_1mm_output_tokens=.6,
+    )
+
 
 class ToolCallShortcut:
     def __init__(self, message: ChatCompletionMessage):
@@ -81,9 +161,9 @@ class ToolCallShortcut:
 
 
 @dataclasses.dataclass
-class CalmResponse:
+class CalmResponse[T: JsonCompatible]:
     success: bool
-    tool_call_results: list[BaseModelOrJsonCompatible]
+    tool_call_results: list[T]
     messages: Messages
     error: Exception | None
     retries_done: int
@@ -120,10 +200,19 @@ class CalmResponse:
     def usage(self) -> CompletionUsage:
         return self._cost_model_usage[2]
 
+    @property
+    def unknown_costs(self) -> bool:
+        """
+        If the model used is not in the cost lookup table, no cost can be determined and this field is True.
+        This can happen with newly released models.
+        """
+        return self._cost_model_usage[3]
+
     @cached_property
-    def _cost_model_usage(self) -> tuple[str | None, float, CompletionUsage]:
+    def _cost_model_usage(self) -> tuple[str | None, float, CompletionUsage, bool]:
         total_cost = 0.
         model = None
+        some_cost_unknown = False
         usage = CompletionUsage(
             completion_tokens=0, prompt_tokens=0, total_tokens=0
         )
@@ -134,33 +223,93 @@ class CalmResponse:
             model = completion.model
             prompt_tokens = completion.usage.prompt_tokens
             completion_tokens = completion.usage.completion_tokens
-            prompt_costs_per_1k, completion_costs_per_1k = COSTS_BY_MODEL[model]
             usage.completion_tokens += completion_tokens
             usage.prompt_tokens += prompt_tokens
-            total_cost += (
-                    prompt_costs_per_1k * prompt_tokens / 1000.
-                    + completion_costs_per_1k * completion_tokens / 1000.
+            prompt_costs_per_1mm, completion_costs_per_1mm = COSTS_BY_MODEL.get(model, (0., 0.))
+            additional_cost = (
+                    prompt_costs_per_1mm * prompt_tokens / 1_000_000.
+                    + completion_costs_per_1mm * completion_tokens / 1_000_000.
             )
+
+            total_cost += additional_cost
+            if additional_cost == 0:
+                some_cost_unknown = True
         usage.total_tokens = usage.prompt_tokens + usage.completion_tokens
-        return model, total_cost, usage
+        return model, total_cost, usage, some_cost_unknown
 
 
-async def get_completion(
+class OpenAIRequestContextManager(Protocol):
+    def __call__(self,
+        model: str,
+        messages: Messages,
+        tools: list[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
+        tool_choice: Literal["none", "auto", "required"] | ChatCompletionNamedToolChoiceParam | NotGiven = NOT_GIVEN,
+        response_format: type[BaseModel] | NotGiven = NOT_GIVEN,
+        **kwargs
+    ) -> AsyncContextManager:
+        ...
+
+@dataclasses.dataclass
+class _OpenAIResultCallbackExc(Exception):
+    """
+    Special Exception used to pass each Completion back to the context manager below
+    """
+    result: ChatCompletion | ParsedChatCompletion
+
+def openai_request_wrapper(func: Callable[..., AsyncGenerator]) -> OpenAIRequestContextManager:
+    """
+    Decorator to wrap an async generator function @asynccontextmanager style to create a wrapper for calls to the OpenAI API.
+    `func` can access the OpenAI completion like so:
+    ```
+    @openai_request_wrapper
+    async def my_openai_wrapper(messages, tools, **kwargs):
+        ... # do something before the request
+        completion: ChatCompletion | ParsedChatCompletion = yield
+        ... # do something after the request / with the response
+    ```
+    The resulting wrapper should be passed to get_completion via the `openai_request_context_manager` param.
+    """
+    @asynccontextmanager
+    async def wrapper(*args, **kwargs):
+        gen = func(*args, **kwargs)
+        async with contextlib.aclosing(gen) as agen:
+            try:
+                res = await agen.asend(None)
+                yield res
+            except _OpenAIResultCallbackExc as e:
+                try:
+                    await agen.asend(e.result)
+                    raise ValueError(
+                        f'Async generator returned from {func.__name__} '
+                        f'did not raise StopAsyncIteration after second call to asend(...)!'
+                    )
+                except StopAsyncIteration:
+                    pass
+
+    wrapper.__is_openai_request_wrapper__ = True
+
+    return wrapper
+
+@openai_request_wrapper
+async def _noop_cm(**kwargs):
+    req = yield
+
+
+async def get_completion[T: JsonCompatible](
         messages: Messages | None = None,
         system_prompt: str | None = None,
         user_message: str | None = None,
-        tools: list[type[BaseModel] | BaseModelFunction] | None = None,
-        tool_choice: Literal[DEFAULT_BEHAVIOR] | ChatCompletionToolChoiceOptionParam = DEFAULT_BEHAVIOR,
+        tools: list[type[T] | JsonCompatibleFunction[T]] | None = None,
+        tool_choice: DefaultBehavior | ChatCompletionToolChoiceOptionParam = DEFAULT_BEHAVIOR,
+        model: Literal["gpt-3.5-turbo", "gpt-4-1106-preview"] | str = None,
+        retries: int = 0,
+        openai_client: AsyncOpenAI | None = None,
         abbreviate_tools: bool = False,
         abbreviation_system_prompt: str | None = "Shortcut tool calling active! If calling tools, omit arguments.",
-        retries: int = 0,
-        rewrite_log_destination: str | TextIO | None = None,
-        rewrite_log_extra_data: dict | None = None,
-        openai_client: AsyncOpenAI | None = None,
-        model: Literal["gpt-3.5-turbo", "gpt-4-1106-preview"] | str = None,
+        openai_request_context_manager: OpenAIRequestContextManager | None = None,
         _track_raw_request_summaries: bool = False,
         **kwargs
-) -> CalmResponse:
+) -> CalmResponse[T]:
     """
     Get a completion with validated function call responses from the chat completions API.
 
@@ -178,16 +327,26 @@ async def get_completion(
         first completion during abbreviated tool calling. Usually this should tell the model not to supply tool
         arguments to not waste tokens.
     :param retries: number of attempts to give the model to fix broken function calls (first try not counted)
-    :param rewrite_log_destination: filename or io handle to log fine-tuning data to
-    :param rewrite_log_extra_data: extra data to merge into the jsonl line for this log entry
-    :param openai_client: optional AsyncOpenAI client to use (use set_client() to set a default client)
+    :param openai_client: optional AsyncOpenAI client to use (use set_openai_client() to set a default client)
     :param model: Which OpenAI model to use for the completion
+    :param openai_request_context_manager: OpenAIRequestContextManager to wrap around every API request
     :param _track_raw_request_summaries: If true, adds a _raw_request_summary field to each of the objects in
         CalmResponse.raw_completions. This can be useful for understanding the full (virtual) message history and set of
         tools that was included with each request.
     :param kwargs: Other keyword arguments to pass to the OpenAI completions API call
     :return: a CalmResponse object
     """
+    if model not in COSTS_BY_MODEL:
+        logging.warning(
+            f"Model {model} is not (yet) known to functioncalming. "
+            f"Cost tracking may be unavailable, and even if structured outputs are supported, they may be deactivated. "
+            "To fix this, call register_model() with the appropriate settings for your model. "
+        )
+
+    openai_request_context_manager = openai_request_context_manager or _noop_cm
+    if not hasattr(openai_request_context_manager, '__is_openai_request_wrapper__'):
+        openai_request_context_manager = openai_request_wrapper(openai_request_context_manager)
+
     # make a copy, we do not edit the passed-in message history
     internal_messages = messages[:] if messages is not None else []
     retries = max(0, retries)
@@ -206,13 +365,17 @@ async def get_completion(
     tools = tools or []
     calm_functions = process_tool_definitions(tools)
 
+    if abbreviate_tools and len(tools) < 2:
+        logging.warning("Abbreviation mode deactivated since there are not multiple tools.")
+        abbreviate_tools = False
+
     abbreviation_mode = abbreviate_tools
     abbreviation_mode_attempted_calls: set[str] = set()
 
     available_function_names: set[str] = set(calm_functions.openai_functions.keys())
 
     # tracks successful tool call outputs
-    result_instances: list[BaseModel] = []
+    result_instances: list[T] = []
 
     # for tracking what "really happened"
     total_completions_generated = 0
@@ -228,6 +391,7 @@ async def get_completion(
         # this will be cut off once abbreviation mode is no longer active
         internal_messages.append({"role": 'system', 'content': abbreviation_system_prompt})
 
+    had_successful_structured_output = False
     errors: list[Exception] | None = []
     while total_completions_generated < total_generations_allowed:
         openai_functions: dict[str, OpenAIFunction] = calm_functions.openai_functions
@@ -240,65 +404,87 @@ async def get_completion(
                 ) from ExceptionGroup("Tool calling validation errors", errors)
             openai_functions = calm_functions.abbreviated_openai_functions
 
-        generated_completion = await _generate_one_completion(
+        generated_completion: ChatCompletion | ParsedChatCompletion = await _generate_one_completion(
             messages=internal_messages,
             openai_functions=openai_functions,
             available_function_names=available_function_names,
             tool_choice=tool_choice,
             model=model,
             openai_client=openai_client,
+            openai_request_context_manager=openai_request_context_manager,
             _track_raw_request_summaries=_track_raw_request_summaries,
             **kwargs
         )
         total_completions_generated += 1
         raw_completions.append(generated_completion)
 
-        last_message: ChatCompletionMessage = generated_completion.choices[0].message
+        last_message: ChatCompletionMessage | ParsedChatCompletionMessage = generated_completion.choices[0].message
+
+        exclusions = set()
+        if not last_message.tool_calls:
+            exclusions.add("tool_calls")  # emtpy list causes a validation error with OpenAI
+        if hasattr(last_message, "parsed") and last_message.parsed is not None:
+            exclusions.add("parsed")  # omit "parsed" in the data sent back to the API
+
         internal_messages.append(
             # make sure the history only has dict objects
-            last_message.model_dump(exclude_unset=True, exclude_none=True)
+            last_message.model_dump(exclude_unset=True, exclude_none=True, exclude=exclusions or None)
         )
 
-        if not last_message.tool_calls:
+        if last_message.tool_calls:
+            # there were tool calls, let's try to execute them
+            outcomes = await execute_tool_calls(
+                tool_calls=last_message.tool_calls,
+                openai_functions=openai_functions
+            )
+            if abbreviation_mode:
+                # track any (valid) function that the model tried to call
+                #  once we exit abbreviation mode, all of them need to be available
+                abbreviation_mode_attempted_calls |= set(
+                    outcome.tool_name for outcome in outcomes if outcome.tool_name is not None
+                )
+
+                # if we are still in abbreviation mode and all calls were successful:
+                #   turn off abbreviation mode
+                #   reset the message history to before the tool calls
+                #   but only allow the tool calls that were actually made
+                if not errors:
+                    # end loop early, cutting all abbreviated function calls from the message history
+                    internal_messages = internal_messages[:rewrite_cutoff]
+                    # TODO omitted_messages is misleading when this code branch is followed
+                    abbreviation_mode = False
+                    # however, also restrict the set of functions to those that the model actually tried to call
+                    available_function_names = abbreviation_mode_attempted_calls
+                    # on the next iteration, the model will now be able to choose only from these functions,
+                    #  but now with full definitions given
+                    continue
+
+            new_successful_instances = [outcome.result for outcome in outcomes if outcome.success]
+            new_successful_tool_calls = [outcome.raw_tool_call for outcome in outcomes if outcome.success]
+            new_successful_tool_responses = [outcome.to_response() for outcome in outcomes if outcome.success]
+            # Note: this may be a mixture of successful responses and errors
+            new_messages = [outcome.to_response() for outcome in outcomes]
+        elif hasattr(last_message, "parsed") and last_message.parsed is not None:
+            assert len(available_function_names) == 1
+            oai_fun_name, = available_function_names
+            oai_fun = openai_functions[oai_fun_name]
+            outcome = await validate_structured_output(last_message, openai_function=oai_fun)
+            had_successful_structured_output = outcome.success
+            outcomes = [outcome]
+            # a structured response is actually not a tool call
+            new_successful_tool_calls = []
+            new_successful_tool_responses = []  # no need to turn these into responses
+            # we only generate a message when the structured output fails
+            new_messages = [outcome.to_response()] if not outcome.success else []
+            new_successful_instances = [outcome.result] if outcome.success else []
+        else:
             # no tool calls: just break the loop (we're done)
             break
 
-        # there were tool calls, let's try to execute them
-        outcomes = await execute_tool_calls(
-            tool_calls=last_message.tool_calls,
-            openai_functions=openai_functions
-        )
-
-        new_successful_instances = [outcome.result for outcome in outcomes if outcome.success]
-        new_successful_tool_calls = [outcome.raw_tool_call for outcome in outcomes if outcome.success]
-        new_successful_tool_responses = [outcome.to_response() for outcome in outcomes if outcome.success]
-        # Note: this may be a mixture of successful responses and errors
-        new_messages = [outcome.to_response() for outcome in outcomes]
-
         # 'errors' is overwritten intentionally, we only ever care about the errors of the last tool call(s)
-        errors = [outcome.result for outcome in outcomes if not outcome.success]
+        errors = [outcome.error for outcome in outcomes if not outcome.success]
 
-        if abbreviation_mode:
-            # track any (valid) function that the model tried to call
-            #  once we exit abbreviation mode, all of them need to be available
-            abbreviation_mode_attempted_calls |= set(
-                outcome.tool_name for outcome in outcomes if outcome.tool_name is not None
-            )
-
-            # if we are still in abbreviation mode and all calls were successful:
-            #   turn off abbreviation mode
-            #   reset the message history to before the tool calls
-            #   but only allow the tool calls that were actually made
-            if not errors:
-                # end loop early, cutting all of the abbreviated function calls from the message history
-                internal_messages = internal_messages[:rewrite_cutoff]
-                abbreviation_mode = False
-                # however, also restrict the set of functions to those that the model actually tried to call
-                available_function_names = abbreviation_mode_attempted_calls
-                # on the next iteration, the model will now be able to choose only from these functions,
-                #  but now with full definitions given
-                continue
-        else:
+        if not abbreviation_mode:
             # if we are not in abbreviation mode, we just track the outputs and go on to handle errors
             result_instances += new_successful_instances
             successful_tool_calls += new_successful_tool_calls
@@ -321,8 +507,8 @@ async def get_completion(
             internal_messages.append({
                 "role": "system",
                 "content": f"{num_failed}/{num_failed + num_successful} tool calls failed. "
-                           f"Please carefully recall the tool definition(s) and repeat all failed calls "
-                           f"(but only repeat the failed calls!)."
+                           f"Please carefully recall the supplied schema definition(s) and try again."
+                           f"If there were multiple calls, only repeat the failed ones!"
             })
             available_function_names = tool_names_for_next_attempt
         else:
@@ -335,20 +521,13 @@ async def get_completion(
         messages=internal_messages,
         rewrite_cutoff=rewrite_cutoff,
         successful_tool_calls=successful_tool_calls,
-        successful_tool_responses=successful_tool_responses
+        successful_tool_responses=successful_tool_responses,
+        had_successful_structured_output=had_successful_structured_output
     )
 
     final_error = None
     if errors:
         final_error = ExceptionGroup("Tool calling validation errors", errors)
-    elif rewrite_log_destination is not None:
-        functions_coalesced: list[OpenAIFunction] = list(calm_functions.openai_functions.values())
-        log_finetuning_data(
-            destination=rewrite_log_destination,
-            messages=internal_messages,
-            functions=functions_coalesced,
-            extra_data=rewrite_log_extra_data
-        )
 
     return CalmResponse(
         success=final_error is None,
@@ -369,22 +548,102 @@ class RawRequestSummary:
     tool_choice: ChatCompletionToolChoiceOptionParam | dict | NotGiven
 
 
+def _maybe_resolve_single_tool_choice(tool_choice_for_api_call, available_function_names):
+    name_of_only_tool = None
+    if tool_choice_for_api_call == "required" and len(available_function_names) == 1:
+        name_of_only_tool, = available_function_names
+    if tool_choice_for_api_call is not NOT_GIVEN and isinstance(tool_choice_for_api_call, dict):
+        tool_choice_for_api_call: ChatCompletionNamedToolChoiceParam
+        name_of_only_tool = tool_choice_for_api_call["function"]["name"]
+    return name_of_only_tool
+
+
+async def _call_openai_with_structured_outputs_if_possible(
+        messages: Messages,
+        model: str,
+        openai_client: AsyncOpenAI,
+        openai_functions: dict[str, OpenAIFunction],
+        available_function_names: set[str],
+        tool_choice_for_api_call: ChatCompletionToolChoiceOptionParam,
+        openai_request_context_manager: OpenAIRequestContextManager,
+        _track_raw_request_summaries: bool,
+        **kwargs
+) -> ChatCompletion | ParsedChatCompletion:
+
+    response_format: type[BaseModel] | None = None
+    name_of_only_tool = _maybe_resolve_single_tool_choice(
+        tool_choice_for_api_call,
+        available_function_names
+    )
+    if name_of_only_tool is not None and openai_functions[name_of_only_tool].was_defined_as_basemodel:
+        # if there is only one tool, and it was defined as a BaseModel, we use response_format
+        # if the tool was a function, we supply it via tools instead - OpenAI docs say that is best practice
+        response_format = openai_functions[name_of_only_tool].non_validating_model
+
+    if response_format and structured_outputs_available(model_name=model):
+        # here we know that we definitely want the output to match one specific JSONSchema spec,
+        # so we can use structured outputs.
+        logging.debug(USING_STRUCTURED_OUTPUTS)
+        async with openai_request_context_manager(
+                messages=messages,
+                model=model,
+                tools=NOT_GIVEN,
+                tool_choice=NOT_GIVEN,
+                response_format=response_format,
+                **kwargs
+        ):
+            generated_completion: ParsedChatCompletion[response_format] = await openai_client.beta.chat.completions.parse(
+                messages=messages,
+                model=model,
+                response_format=response_format
+            )
+            raise _OpenAIResultCallbackExc(result=generated_completion)
+        tools_for_api_call = [openai_functions[name_of_only_tool].tool_definition] or NOT_GIVEN
+    else:
+        tools_for_api_call = [
+            openai_functions[name].tool_definition
+            for name
+            in available_function_names
+        ] or NOT_GIVEN
+        async with openai_request_context_manager(
+            messages=messages,
+            model=model,
+            tools=tools_for_api_call,
+            tool_choice=tool_choice_for_api_call,
+            response_format=NOT_GIVEN,
+            **kwargs
+        ):
+            generated_completion: ChatCompletion = await openai_client.chat.completions.create(
+                messages=messages,
+                model=model,
+                tools=tools_for_api_call,
+                tool_choice=tool_choice_for_api_call,
+                **kwargs
+            )
+            raise _OpenAIResultCallbackExc(result=generated_completion)
+    if _track_raw_request_summaries:
+        generated_completion._raw_request_summary = RawRequestSummary(
+            messages=messages[:], tools=tools_for_api_call, tool_choice=tool_choice_for_api_call
+        )
+    return generated_completion
+
+
 async def _generate_one_completion(
         messages: Messages,
         openai_functions: dict[str, OpenAIFunction],
         available_function_names: set[str],
-        tool_choice: Literal[DEFAULT_BEHAVIOR] | ChatCompletionToolChoiceOptionParam,
+        tool_choice: DefaultBehavior | ChatCompletionToolChoiceOptionParam,
         model: str,  # todo
         openai_client: AsyncOpenAI,
+        openai_request_context_manager: OpenAIRequestContextManager,
         _track_raw_request_summaries: bool,
         **kwargs
-):
+) -> ChatCompletion | ParsedChatCompletion:
+    tool_choice_for_api_call: ChatCompletionToolChoiceOptionParam
     if tool_choice == DEFAULT_BEHAVIOR:
-        current_tool_choice = get_tool_choice_for_default_behavior(available_function_names)
+        tool_choice_for_api_call = get_tool_choice_for_default_behavior(available_function_names)
     else:
-        current_tool_choice = tool_choice
-
-    current_tools = [openai_functions[name].tool_definition for name in available_function_names] or NOT_GIVEN
+        tool_choice_for_api_call = tool_choice
 
     shortcut: ChatCompletionMessage | None = await maybe_shortcut_trivial_function_call(
         available_function_names,
@@ -394,53 +653,97 @@ async def _generate_one_completion(
     if shortcut is not None:
         generated_completion = ToolCallShortcut(message=shortcut)
     else:
-        generated_completion = await openai_client.chat.completions.create(
+        generated_completion = await _call_openai_with_structured_outputs_if_possible(
             messages=messages,
             model=model,
-            tools=current_tools,
-            tool_choice=current_tool_choice,
+            openai_client=openai_client,
+            openai_functions=openai_functions,
+            available_function_names=available_function_names,
+            tool_choice_for_api_call=tool_choice_for_api_call,
+            openai_request_context_manager=openai_request_context_manager,
+            _track_raw_request_summaries=_track_raw_request_summaries,
             **kwargs
         )
 
-    if _track_raw_request_summaries:
-        generated_completion._raw_request_summary = RawRequestSummary(
-            messages=messages[:], tools=current_tools, tool_choice=current_tool_choice
-        )
     return generated_completion
 
 
+@dataclasses.dataclass
+class StructuredOutputOutcome:
+    success: bool
+    raw_content: str
+    result: JsonCompatible
+    error: BaseException | None
+    tool_name: str | None
+
+    def to_response(self) -> ChatCompletionSystemMessageParam:
+        if self.success:
+            raise ValueError("Shouldn't need to call to_response on a successful structured response.")
+        return ChatCompletionSystemMessageParam(
+            role="system",
+            content=f"Error: {serialize_openai_function_result(self.result)}"
+        )
+
 @dataclasses.dataclass
 class ToolCallOutcome:
     success: bool
     tool_call_id: str
     raw_tool_call: ChatCompletionMessageToolCall
-    result: BaseModelOrJsonCompatible | Exception
+    result: JsonCompatible
+    error: BaseException | None
     tool_name: str | None
 
     def to_response(self) -> ChatCompletionToolMessageParam:
         return ChatCompletionToolMessageParam(
             role="tool",
             tool_call_id=self.tool_call_id,
-            content=serialize_openai_function_result(self.result) if self.success else f"Error: {self.result}"
+            content=serialize_openai_function_result(self.result)
         )
 
 
+async def validate_structured_output(
+        message: ParsedChatCompletionMessage, openai_function: OpenAIFunction
+) -> StructuredOutputOutcome:
+    success = True
+    error = None
+    try:
+        result = await openai_function.callback(**message.parsed.model_dump())
+    except ValidationError as e:
+        error = e
+        result = e.json()
+        success = False
+    except Exception as e:
+        error = e
+        result = f"Error: {e}"
+        success = False
+
+    return StructuredOutputOutcome(
+        success=success,
+        raw_content=message.content,
+        result=result,
+        error=error,
+        tool_name=openai_function.name
+    )
+
 async def execute_tool_calls(
-        tool_calls: list[ChatCompletionMessageToolCall], openai_functions
+        tool_calls: list[ChatCompletionMessageToolCall],
+        openai_functions: dict[str, OpenAIFunction]
 ) -> list[ToolCallOutcome]:
     outcomes = []
     for tool_call in tool_calls:
         function_name = tool_call.function.name
 
         if function_name not in openai_functions:
-            e = ToolCallError(f"Error: function `{function_name}` does not exist.")
+            msg = f"Error: function `{function_name}` does not exist."
+            e = ToolCallError(msg)
 
             outcomes.append(
                 ToolCallOutcome(
                     success=False,
                     tool_call_id=tool_call.id,
                     raw_tool_call=tool_call,
-                    result=e,
+                    result=msg,
+                    error=e,
                     tool_name=None
                 )
             )
@@ -455,20 +758,26 @@ async def execute_tool_calls(
                     success=False,
                     tool_call_id=tool_call.id,
                     raw_tool_call=tool_call,
-                    result=e,
+                    result=f'Error: invalid JSON ({e})',
+                    error=e,
                     tool_name=function_name
                 )
             )
             continue
 
         try:
-            result_instance = await openai_function.callback(**arguments)
+            with set_calm_context(
+                tool_call=tool_call,
+                openai_function=openai_function,
+            ):
+                result_instance = await openai_function.callback(**arguments)
             outcomes.append(
                 ToolCallOutcome(
                     success=True,
                     tool_call_id=tool_call.id,
                     raw_tool_call=tool_call,
                     result=result_instance,
+                    error=None,
                     tool_name=function_name
                 )
             )
@@ -478,13 +787,25 @@ async def execute_tool_calls(
             # since the model did fine calling the tool here.
             # To force a retry due to a semantic error the model should correct, raise a ToolCallError instead
             raise e
-        except (ValidationError, ToolCallError) as e:
+        except ValidationError as e:
             outcomes.append(
                 ToolCallOutcome(
                     success=False,
                     tool_call_id=tool_call.id,
                     raw_tool_call=tool_call,
-                    result=e,
+                    result=e.json(),
+                    error=e,
+                    tool_name=function_name
+                )
+            )
+        except ToolCallError as e:
+            outcomes.append(
+                ToolCallOutcome(
+                    success=False,
+                    tool_call_id=tool_call.id,
+                    raw_tool_call=tool_call,
+                    result=f'Tool call error: {e}',
+                    error=e,
                     tool_name=function_name
                 )
             )
@@ -560,12 +881,14 @@ def process_tool_definitions(tools_raw) -> ProcessedFunctions:
     # Note: these are indexed by their original name
     abbreviated_functions: dict[str, OpenAIFunction] = {}
     for model_or_fn in tools_raw:
-        openai_function = create_openai_function(model_or_fn)
+        openai_function = create_openai_function(model_or_fn, keep_in_model_cache=True)
         openai_functions[openai_function.name] = openai_function
 
         abbreviated_function = create_abbreviated_openai_function(model_or_fn=model_or_fn)
         abbreviated_functions[abbreviated_function.name] = abbreviated_function
 
+    rebuild_cached_models()
+
     return ProcessedFunctions(
         openai_functions=openai_functions, abbreviated_openai_functions=abbreviated_functions
     )
@@ -590,13 +913,19 @@ def rewrite_message_history(
         rewrite_cutoff: int,
         successful_tool_calls: list[ChatCompletionMessageToolCall],
         successful_tool_responses: list[ChatCompletionToolMessageParam],
+        had_successful_structured_output: bool
 ):
     """
     Rewrite the history by removing all messages (not just the failed calls) and replacing them with a single clean
     multi call and its responses, starting from the cutoff index.
     """
-    omitted_messages = messages[rewrite_cutoff:]
+    omitted_messages = []
+    stashed_final_structured_output = None
+    if had_successful_structured_output:
+        stashed_final_structured_output = messages[-1]
+
     if successful_tool_calls:
+        omitted_messages = messages[rewrite_cutoff:-1 if stashed_final_structured_output else None]
         messages[rewrite_cutoff:] = [
             {
                 "role": "assistant",
@@ -605,5 +934,11 @@ def rewrite_message_history(
             },
             *successful_tool_responses
         ]
+    elif had_successful_structured_output:
+        omitted_messages = messages[rewrite_cutoff:-1]
+        messages[:] = messages[:rewrite_cutoff]
+
+    if stashed_final_structured_output:
+        messages.append(stashed_final_structured_output)
     return omitted_messages
 
